@@ -103,8 +103,8 @@ static void                 shutdown_databases (void);
 static void                 set_default_database (GConfDatabase* db);
 static void                 register_database (GConfDatabase* db);
 static void                 unregister_database (GConfDatabase* db);
-static GConfDatabase*       lookup_database (const gchar *address);
-static GConfDatabase*       obtain_database (const gchar *address,
+static GConfDatabase*       lookup_database (GSList *addresses);
+static GConfDatabase*       obtain_database (GSList *addresses,
                                              GError **err);
 static void                 drop_old_databases (void);
 static gboolean             no_databases_in_use (void);
@@ -120,11 +120,17 @@ static void gconf_handle_segv (int signum);
 
 static gboolean in_shutdown = FALSE;
 
+/*
+ * Flag indicating we received a SIGHUP and we should reaload
+ * all sources during the next periodic_cleanup()
+ */
+static gboolean need_db_reload = FALSE;
+
 /* 
  * CORBA goo
  */
 
-static ConfigServer server = CORBA_OBJECT_NIL;
+static ConfigServer2 server = CORBA_OBJECT_NIL;
 static PortableServer_POA the_poa;
 static GConfLock *daemon_lock = NULL;
 
@@ -137,6 +143,11 @@ gconfd_get_database(PortableServer_Servant servant,
                     const CORBA_char* address,
                     CORBA_Environment* ev);
 
+static ConfigDatabase
+gconfd_get_database_for_addresses (PortableServer_Servant           servant,
+				   const ConfigServer2_AddressList *addresses,
+				   CORBA_Environment               *ev);
+
 static void
 gconfd_add_client (PortableServer_Servant servant,
                    const ConfigListener client,
@@ -169,8 +180,13 @@ static POA_ConfigServer__epv server_epv = {
   gconfd_shutdown
 };
 
-static POA_ConfigServer__vepv poa_server_vepv = { &base_epv, &server_epv };
-static POA_ConfigServer poa_server_servant = { NULL, &poa_server_vepv };
+static POA_ConfigServer2__epv server2_epv = { 
+  NULL,
+  gconfd_get_database_for_addresses
+};
+
+static POA_ConfigServer2__vepv poa_server_vepv = { &base_epv, &server_epv, &server2_epv };
+static POA_ConfigServer2 poa_server_servant = { NULL, &poa_server_vepv };
 
 static ConfigDatabase
 gconfd_get_default_database(PortableServer_Servant servant,
@@ -195,19 +211,51 @@ gconfd_get_database(PortableServer_Servant servant,
                     CORBA_Environment* ev)
 {
   GConfDatabase *db;
+  GSList *addresses;
   GError* error = NULL;  
 
   if (gconfd_check_in_shutdown (ev))
     return CORBA_OBJECT_NIL;
   
-  db = obtain_database (address, &error);
+  addresses = g_slist_append (NULL, (char *) address);
+  db = obtain_database (addresses, &error);
+  g_slist_free (addresses);
 
   if (db != NULL)
     return CORBA_Object_duplicate (db->objref, ev);
-  else if (gconf_set_exception(&error, ev))
-    return CORBA_OBJECT_NIL;
-  else
+
+  gconf_set_exception (&error, ev);
+
+  return CORBA_OBJECT_NIL;
+}
+
+static ConfigDatabase
+gconfd_get_database_for_addresses (PortableServer_Servant           servant,
+				   const ConfigServer2_AddressList *seq,
+				   CORBA_Environment               *ev)
+{
+  GConfDatabase  *db;
+  GSList         *addresses = NULL;
+  GError         *error = NULL;  
+  int             i;
+
+  if (gconfd_check_in_shutdown (ev))
     return CORBA_OBJECT_NIL;
+
+  i = 0;
+  while (i < seq->_length)
+    addresses = g_slist_append (addresses, seq->_buffer [i++]);
+
+  db = obtain_database (addresses, &error);
+
+  g_slist_free (addresses);
+
+  if (db != NULL)
+    return CORBA_Object_duplicate (db->objref, ev);
+
+  gconf_set_exception (&error, ev);
+
+  return CORBA_OBJECT_NIL;
 }
 
 static void
@@ -301,7 +349,7 @@ gconf_server_load_sources(void)
       /* We want to stay alive but do nothing, because otherwise every
          request would result in another failed gconfd being spawned.  
       */
-      gconf_log(GCL_ERR, _("No configuration sources in the source path, configuration won't be saved; edit "GCONF_CONFDIR"/path"));
+      gconf_log(GCL_ERR, _("No configuration sources in the source path, configuration won't be saved; edit %s"), GCONF_CONFDIR"/path");
       /* don't request error since there aren't any addresses */
       sources = gconf_sources_new_from_addresses(NULL, NULL);
 
@@ -397,7 +445,6 @@ signal_handler (int signo)
     break;
 
   case SIGTERM:
-  case SIGHUP:
     enter_shutdown ();
 
     /* let the fatal signals interrupt us */
@@ -410,6 +457,13 @@ signal_handler (int signo)
       gconf_main_quit ();
     break;
 
+  case SIGHUP:
+    --in_fatal;
+
+    /* reload sources during next periodic_cleanup() */
+    need_db_reload = TRUE;
+    break;
+
   case SIGUSR1:
     --in_fatal;
     
@@ -423,7 +477,7 @@ signal_handler (int signo)
 }
 
 PortableServer_POA
-gconf_get_poa ()
+gconf_get_poa (void)
 {
   return the_poa;
 }
@@ -468,7 +522,6 @@ log_handler (const gchar   *log_domain,
  * the given directory and keeps us from creating
  * it
  */
-/* FIXME this is just paranoia now since we use linc_get_tmpdir() */
 static gboolean
 test_safe_tmp_dir (const char *dirname)
 {
@@ -502,8 +555,8 @@ test_safe_tmp_dir (const char *dirname)
   if ((statbuf.st_mode & (S_IRWXG|S_IRWXO)) ||
       !S_ISDIR (statbuf.st_mode))
     {
-      gconf_log (GCL_WARNING, _("Bad permissions %o on dir %s"),
-                 statbuf.st_mode & 07777, dirname);
+      gconf_log (GCL_WARNING, _("Bad permissions %lo on directory %s"),
+                 (unsigned long) statbuf.st_mode & 07777, dirname);
       return FALSE;
     }
   
@@ -519,7 +572,6 @@ main(int argc, char** argv)
   CORBA_ORB orb;
   gchar* logname;
   const gchar* username;
-  guint len;
   gchar* ior;
   int exit_code = 0;
   GError *err;
@@ -543,7 +595,7 @@ main(int argc, char** argv)
    */
   if (chdir ("/") < 0)
     {
-       fprintf (stderr, "Could not change to root directory: %s\n",
+       g_printerr ("Could not change to root directory: %s\n",
 		g_strerror (errno));
        exit (1);
     }
@@ -569,9 +621,7 @@ main(int argc, char** argv)
   
   /* Logs */
   username = g_get_user_name();
-  len = strlen(username) + strlen("gconfd") + 15;
-  logname = g_malloc(len);
-  g_snprintf(logname, len, "gconfd (%s-%u)", username, (guint)getpid());
+  logname = g_strdup_printf("gconfd (%s-%u)", username, (guint)getpid());
 
   openlog (logname, LOG_NDELAY, LOG_USER);
 
@@ -617,7 +667,7 @@ main(int argc, char** argv)
 
   orb = gconf_orb_get ();
   
-  POA_ConfigServer__init (&poa_server_servant, &ev);
+  POA_ConfigServer2__init (&poa_server_servant, &ev);
   
   the_poa = (PortableServer_POA)CORBA_ORB_resolve_initial_references(orb, "RootPOA", &ev);
   PortableServer_POAManager_activate(PortableServer_POA__get_the_POAManager(the_poa, &ev), &ev);
@@ -751,6 +801,16 @@ static gboolean need_log_cleanup = FALSE;
 static gboolean
 periodic_cleanup_timeout(gpointer data)
 {  
+  if (need_db_reload)
+    {
+      gconf_log (GCL_INFO, _("SIGHUP received, reloading all databases"));
+
+      need_db_reload = FALSE;
+      shutdown_databases ();
+      init_databases ();
+      gconf_server_load_sources ();		  
+    }
+  
   gconf_log (GCL_DEBUG, "Performing periodic cleanup, expiring cache cruft");
   
   drop_old_clients ();
@@ -795,7 +855,7 @@ gconf_main(void)
 
   if (main_loops == NULL)
     {
-      gulong timeout_len = 1000*60*2; /* 1 sec * 60 s/min * 2 min */
+      gulong timeout_len = 1000*60*0.5; /* 1 sec * 60 s/min * .5 min */
       
       g_assert(timeout_id == 0);
       timeout_id = g_timeout_add (timeout_len,
@@ -839,7 +899,7 @@ gconf_main_is_running (void)
  */
 
 static GList* db_list = NULL;
-static GHashTable* dbs_by_address = NULL;
+static GHashTable* dbs_by_addresses = NULL;
 static GConfDatabase *default_db = NULL;
 
 static void
@@ -848,13 +908,9 @@ init_databases (void)
   gconfd_need_log_cleanup ();
   
   g_assert(db_list == NULL);
-  g_assert(dbs_by_address == NULL);
+  g_assert(dbs_by_addresses == NULL);
   
-  dbs_by_address = g_hash_table_new (g_str_hash, g_str_equal);
-
-  /* Default database isn't in the address hash since it has
-     multiple addresses in a stack
-  */
+  dbs_by_addresses = g_hash_table_new (g_str_hash, g_str_equal);
 }
 
 static void
@@ -863,10 +919,8 @@ set_default_database (GConfDatabase* db)
   gconfd_need_log_cleanup ();
   
   default_db = db;
-  
-  /* Default database isn't in the address hash since it has
-     multiple addresses in a stack
-  */
+
+  register_database (db);
 }
 
 static void
@@ -875,9 +929,9 @@ register_database (GConfDatabase *db)
   gconfd_need_log_cleanup ();
   
   if (db->sources->sources)
-    safe_g_hash_table_insert(dbs_by_address,
-                             ((GConfSource*)db->sources->sources->data)->address,
-                             db);
+    safe_g_hash_table_insert (dbs_by_addresses,
+			      (char *) gconf_database_get_persistent_name (db),
+			      db);
   
   db_list = g_list_prepend (db_list, db);
 }
@@ -888,8 +942,10 @@ unregister_database (GConfDatabase *db)
   gconfd_need_log_cleanup ();
   
   if (db->sources->sources)
-    g_hash_table_remove(dbs_by_address,
-                        ((GConfSource*)(db->sources->sources->data))->address);
+    {
+      g_hash_table_remove (dbs_by_addresses,
+			   gconf_database_get_persistent_name (db));
+    }
 
   db_list = g_list_remove (db_list, db);
 
@@ -897,32 +953,37 @@ unregister_database (GConfDatabase *db)
 }
 
 static GConfDatabase*
-lookup_database (const gchar *address)
+lookup_database (GSList *addresses)
 {
-  if (address == NULL)
+  GConfDatabase *retval;
+  char          *key;
+
+  if (addresses == NULL)
     return default_db;
-  else
-    return g_hash_table_lookup (dbs_by_address, address);
+
+  key = gconf_address_list_get_persistent_name (addresses);
+
+  retval = g_hash_table_lookup (dbs_by_addresses, key);
+
+  g_free (key);
+
+  return retval;
 }
 
 static GConfDatabase*
-obtain_database (const gchar *address,
+obtain_database (GSList  *addresses,
                  GError **err)
 {
-  
   GConfSources* sources;
-  GSList* addresses = NULL;
   GError* error = NULL;
   GConfDatabase *db;
 
-  db = lookup_database (address);
+  db = lookup_database (addresses);
 
   if (db)
     return db;
 
-  addresses = g_slist_append(addresses, g_strdup(address));
   sources = gconf_sources_new_from_addresses(addresses, &error);
-  g_slist_free (addresses);
 
   if (error != NULL)
     {
@@ -1009,14 +1070,10 @@ shutdown_databases (void)
   g_list_free (db_list);
   db_list = NULL;
 
-  if (dbs_by_address)
-    g_hash_table_destroy(dbs_by_address);
-
-  dbs_by_address = NULL;
-
-  if (default_db)
-    gconf_database_free (default_db);
+  if (dbs_by_addresses)
+    g_hash_table_destroy(dbs_by_addresses);
 
+  dbs_by_addresses = NULL;
   default_db = NULL;
 }
 
@@ -1030,6 +1087,76 @@ no_databases_in_use (void)
     gconf_listeners_count (default_db->listeners) == 0;
 }
 
+void
+gconfd_notify_other_listeners (GConfDatabase *modified_db,
+			       GConfSources  *modified_sources,
+                               const char    *key)
+{
+  GList *tmp;
+
+  if (!modified_sources)
+    return;
+  
+  tmp = db_list;
+  while (tmp != NULL)
+    {
+      GConfDatabase *db = tmp->data;
+
+      if (db != modified_db)
+	{
+	  GList *tmp2;
+
+	  tmp2 = modified_sources->sources;
+	  while (tmp2)
+	    {
+	      GConfSource *modified_source = tmp2->data;
+
+	      if (gconf_sources_is_affected (db->sources, modified_source, key))
+		{
+		  GConfValue  *value;
+		  ConfigValue *cvalue;
+		  GError      *error;
+		  gboolean     is_default;
+		  gboolean     is_writable;
+
+		  error = NULL;
+		  value = gconf_database_query_value (db,
+						      key,
+						      NULL,
+						      TRUE,
+						      NULL,
+						      &is_default,
+						      &is_writable,
+						      &error);
+		  if (error != NULL)
+		    {
+		      gconf_log (GCL_WARNING,
+				 _("Error obtaining new value for `%s': %s"),
+				 key, error->message);
+		      g_error_free (error);
+		      return;
+		    }
+
+		  cvalue = gconf_corba_value_from_gconf_value (value);
+		  gconf_database_notify_listeners (db,
+						   NULL,
+						   key,
+						   cvalue,
+						   is_default,
+						   is_writable,
+						   FALSE);
+		  CORBA_free (cvalue);
+		  gconf_value_free (value);
+		}
+
+	      tmp2 = tmp2->next;
+	    }
+	}
+
+      tmp = tmp->next;
+    }
+}
+
 /*
  * Cleanup
  */
@@ -1295,26 +1422,19 @@ logfile_save (void)
                          * that matter on open()
                          */
 
-  saveme = g_string_new ("");
+  saveme = g_string_new (NULL);
 
   /* Clients */
   log_clients_to_string (saveme);
   
-  /* Default database */
-  gconf_database_log_listeners_to_string (default_db,
-                                          TRUE,
-                                          saveme);
-
-  /* Other databases */
-  
+  /* Databases */
   tmp_list = db_list;
-
   while (tmp_list)
     {
       GConfDatabase *db = tmp_list->data;
 
       gconf_database_log_listeners_to_string (db,
-                                              FALSE,
+                                              db == default_db ? TRUE : FALSE,
                                               saveme);
       
       tmp_list = g_list_next (tmp_list);
@@ -1845,12 +1965,20 @@ listener_logentry_restore_and_destroy_foreach (gpointer key,
                                                gpointer data)
 {
   ListenerLogEntry *lle = key;
-  GConfDatabase *db;
+  GConfDatabase *db = NULL;
   
   if (strcmp (lle->address, "def") == 0)
     db = default_db;
   else
-    db = obtain_database (lle->address, NULL);
+    {
+      GSList *addresses;
+
+      addresses = gconf_persistent_name_get_address_list (lle->address);
+
+      db = obtain_database (addresses, NULL);
+
+      gconf_address_list_free (addresses);
+    }
   
   if (db == NULL)
     {
@@ -1887,7 +2015,7 @@ read_line (FILE *f)
   int c;
   GString *str;
   
-  str = g_string_new ("");
+  str = g_string_new (NULL);
   
   flockfile (f);
 
@@ -1954,8 +2082,9 @@ logfile_read (void)
   
   if (f == NULL)
     {
-      gconf_log (GCL_ERR, _("Unable to open saved state file '%s': %s"),
-                 logfile, g_strerror (errno));
+      if (errno != ENOENT)
+          gconf_log (GCL_ERR, _("Unable to open saved state file '%s': %s"),
+                     logfile, g_strerror (errno));
 
       goto finished;
     }
