@@ -29,8 +29,6 @@
 #include <stdlib.h>
 #include <errno.h>
 #include <locale.h>
-#include <sys/types.h>
-#include <sys/stat.h>
 
 typedef enum {
   LOAD_SCHEMA_FILE,
@@ -61,6 +59,7 @@ static char* long_desc = NULL;
 static char* owner = NULL;
 static char* schema_file = NULL;
 static char* entry_file = NULL;
+static char* unload_entry_file = NULL;
 static const char* config_source = NULL;
 static int use_local_source = FALSE;
 static int makefile_install_mode = FALSE;
@@ -72,6 +71,7 @@ static int long_docs_mode = FALSE;
 static int schema_name_mode = FALSE;
 static int associate_schema_mode = FALSE;
 static int dissociate_schema_mode = FALSE;
+static int ignore_schema_defaults = FALSE;
 static int default_source_mode = FALSE;
 static int recursive_unset_mode = FALSE;
 static int do_version = FALSE;
@@ -168,6 +168,15 @@ struct poptOption options[] = {
     N_("Load from the specified file an XML description of values and set them relative to a dir."),
     NULL
   },
+  {
+    "unload",
+    '\0',
+    POPT_ARG_STRING,
+    &unload_entry_file,
+    0,
+    N_("Unload a set of values described in an XML file."),
+    NULL
+  },
   {
     "recursive-list",
     'R',
@@ -411,6 +420,15 @@ struct poptOption options[] = {
     N_("Remove any schema name applied to the given keys"),
     NULL
   },
+  {
+    "ignore-schema-defaults",
+    '\0',
+    POPT_ARG_NONE,
+    &ignore_schema_defaults,
+    0,
+    N_("Ignore schema defaults when reading values."),
+    NULL
+  },
   {
     "get-default-source",
     '\0',
@@ -442,8 +460,7 @@ struct poptOption options[] = {
 
 static int do_break_key(GConfEngine* conf, const gchar** args);
 static int do_break_directory(GConfEngine* conf, const gchar** args);
-static int do_makefile_install(GConfEngine* conf, const gchar** args);
-static int do_makefile_uninstall(GConfEngine* conf, const gchar** args);
+static int do_makefile_install(GConfEngine* conf, const gchar** args, gboolean unload);
 static int do_recursive_list(GConfEngine* conf, const gchar** args);
 static int do_dump_values(GConfEngine* conf, const gchar** args);
 static int do_all_pairs(GConfEngine* conf, const gchar** args);
@@ -464,7 +481,7 @@ static int do_all_entries(GConfEngine* conf, const gchar** args);
 static int do_unset(GConfEngine* conf, const gchar** args);
 static int do_recursive_unset (GConfEngine* conf, const gchar** args);
 static int do_all_subdirs(GConfEngine* conf, const gchar** args);
-static int do_load_file(GConfEngine* conf, LoadType load_type, const gchar* file, const gchar** base_dirs);
+static int do_load_file(GConfEngine* conf, LoadType load_type, gboolean unload, const gchar* file, const gchar** base_dirs);
 static int do_short_docs (GConfEngine *conf, const gchar **args);
 static int do_long_docs (GConfEngine *conf, const gchar **args);
 static int do_get_schema_name (GConfEngine *conf, const gchar **args);
@@ -590,7 +607,7 @@ main (int argc, char** argv)
     {
       g_printerr (_("--set_schema should not be used with --get, --set, --unset, --all-entries, --all-dirs\n"));
       return 1;
-    }  
+    }
 
   if ((value_type != NULL) && !(set_mode || set_schema_mode))
     {
@@ -604,6 +621,14 @@ main (int argc, char** argv)
       return 1;
     }
 
+  if (ignore_schema_defaults && !(get_mode || all_entries_mode ||
+				  dump_values || recursive_list ||
+				  get_list_size_mode || get_list_element_mode))
+    {
+      g_printerr (_("--ignore-schema-defaults is only relevant with --get, --all-entries, --dump, --recursive-list, --get-list-size or --get-list-element\n"));
+      return 1;
+    }
+
   if (ping_gconfd && (shutdown_gconfd || set_mode || get_mode || unset_mode ||
                       all_subdirs_mode || all_entries_mode || recursive_list || 
                       get_type_mode || get_list_size_mode || get_list_element_mode ||
@@ -705,12 +730,6 @@ main (int argc, char** argv)
       return 1;
     }
 
-  if (config_source && !use_local_source)
-    {
-      g_printerr (_("You should use --direct when using a non-default configuration source\n"));
-      return 1;
-    }
-  
   if (!gconf_init(argc, argv, &err))
     {
       g_printerr (_("Failed to init GConf: %s\n"), err->message);
@@ -782,10 +801,16 @@ main (int argc, char** argv)
     conf = gconf_engine_get_default();
   else
     {
+      GSList *addresses;
+
+      addresses = gconf_persistent_name_get_address_list (config_source);
+
       if (use_local_source)
-        conf = gconf_engine_get_local(config_source, &err);
+        conf = gconf_engine_get_local_for_addresses (addresses, &err);
       else
-        conf = gconf_engine_get_for_address(config_source, &err);
+        conf = gconf_engine_get_for_addresses (addresses, &err);
+
+      gconf_address_list_free (addresses);
     }
   
   if (conf == NULL)
@@ -821,7 +846,7 @@ main (int argc, char** argv)
     {
       gint retval;
 
-      retval = do_load_file(conf, LOAD_SCHEMA_FILE, schema_file, NULL);
+      retval = do_load_file(conf, LOAD_SCHEMA_FILE, FALSE, schema_file, NULL);
 
       gconf_engine_unref(conf);
 
@@ -833,13 +858,25 @@ main (int argc, char** argv)
       const gchar** args = poptGetArgs(ctx);
       gint retval;
 
-      retval = do_load_file(conf, LOAD_ENTRY_FILE, entry_file, args);
+      retval = do_load_file(conf, LOAD_ENTRY_FILE, FALSE, entry_file, args);
 
       gconf_engine_unref(conf);
 
       return retval;
     }
   
+  if (unload_entry_file != NULL)
+    {
+      const gchar** args = poptGetArgs(ctx);
+      gint retval;
+
+      retval = do_load_file(conf, LOAD_ENTRY_FILE, TRUE, unload_entry_file, args);
+
+      gconf_engine_unref(conf);
+
+      return retval;
+    }
+
   if (spawn_gconfd)
     {
       do_spawn_daemon(conf);
@@ -850,10 +887,7 @@ main (int argc, char** argv)
   if (makefile_install_mode)
     {
       const gchar** args = poptGetArgs(ctx);
-      gint retval;
-
-      umask (022);
-      retval = do_makefile_install (conf, args);
+      gint retval = do_makefile_install (conf, args, FALSE);
       
       gconf_engine_unref (conf);
 
@@ -863,10 +897,7 @@ main (int argc, char** argv)
   if (makefile_uninstall_mode)
     {
       const gchar** args = poptGetArgs(ctx);
-      gint retval;
-
-      umask (022);
-      retval = do_makefile_uninstall (conf, args);
+      gint retval = do_makefile_install (conf, args, TRUE);
       
       gconf_engine_unref (conf);
 
@@ -1234,7 +1265,8 @@ list_pairs_in_dir(GConfEngine* conf, const gchar* dir, guint depth)
           GConfEntry* pair = tmp->data;
           gchar* s;
 
-          if (gconf_entry_get_value (pair))
+          if (gconf_entry_get_value (pair) && 
+	      (!ignore_schema_defaults || !gconf_entry_get_is_default (pair)))
             s = gconf_value_to_string (gconf_entry_get_value (pair));
           else
             s = g_strdup(_("(no value set)"));
@@ -1505,7 +1537,8 @@ dump_entries_in_dir(GConfEngine* conf, const gchar* dir, const gchar* base_dir)
         g_print ("      <schema_key>%s</schema_key>\n",
 		 get_key_relative(gconf_entry_get_schema_name(entry), base_dir));
 
-      if (entry->value)
+      if (entry->value && 
+	  (!ignore_schema_defaults || !gconf_entry_get_is_default(entry)))
         print_value_in_xml(entry->value, 6);
 
       g_print ("    </entry>\n");
@@ -1549,6 +1582,21 @@ do_spawn_daemon(GConfEngine* conf)
     }
 }
 
+static inline GConfValue *
+get_maybe_without_default (GConfEngine  *conf,
+			   const gchar  *key,
+			   GError      **err)
+{
+  if (!ignore_schema_defaults)
+    {
+      return gconf_engine_get (conf, key, err);
+    }
+  else
+    {
+      return gconf_engine_get_without_default (conf, key, err);
+    }
+}
+
 static int
 do_get(GConfEngine* conf, const gchar** args)
 {
@@ -1567,7 +1615,7 @@ do_get(GConfEngine* conf, const gchar** args)
 
       err = NULL;
 
-      value = gconf_engine_get (conf, *args, &err);
+      value = get_maybe_without_default (conf, *args, &err);
          
       if (value != NULL)
         {
@@ -1826,7 +1874,7 @@ do_get_type(GConfEngine* conf, const gchar** args)
 
       err = NULL;
 
-      value = gconf_engine_get (conf, *args, &err);
+      value = get_maybe_without_default (conf, *args, &err);
          
       if (value != NULL)
 	{
@@ -1864,7 +1912,7 @@ do_get_list_size(GConfEngine* conf, const gchar** args)
       return 1;
     }
 
-  list = gconf_engine_get (conf, *args, &err);
+  list = get_maybe_without_default (conf, *args, &err);
 
   if (list == NULL)
     {
@@ -1909,7 +1957,7 @@ do_get_list_element(GConfEngine* conf, const gchar** args)
       return 1;
     }
 
-  list = gconf_engine_get (conf, *args, &err);
+  list = get_maybe_without_default (conf, *args, &err);
 
   if (list == NULL)
     {
@@ -2767,7 +2815,7 @@ typedef struct {
 } EntryInfo;
 
 static void
-set_values(GConfEngine* conf, const gchar* base_dir, const gchar* key, const char* schema_key, GSList* values)
+set_values(GConfEngine* conf, gboolean unload, const gchar* base_dir, const gchar* key, const char* schema_key, GSList* values)
 {
   GSList* tmp;
   gchar* full_key;
@@ -2777,30 +2825,29 @@ set_values(GConfEngine* conf, const gchar* base_dir, const gchar* key, const cha
   else
     full_key = g_strdup(key);
 
-  if (schema_key)
+  if (unload || schema_key)
     {
-      gchar* full_schema_key;
+      gchar* full_schema_key = NULL;
+      GError* error = NULL;
 
-      if (base_dir && *schema_key != '/')
-        full_schema_key = gconf_concat_dir_and_key(base_dir, schema_key);
-      else
-        full_schema_key = g_strdup(schema_key);
-
-      if (full_schema_key)
+      if (!unload)
         {
-	  GError* error = NULL;
-	  
-          if (!gconf_engine_associate_schema(conf, full_key, full_schema_key,  &error))
-            {
-              g_assert(error != NULL);
-          
-              g_printerr (_("WARNING: failed to associate schema `%s' with key `%s': %s\n"),
-			  full_schema_key, full_key, error->message);
-              g_error_free(error);
-            }
+          if (base_dir && *schema_key != '/')
+            full_schema_key = gconf_concat_dir_and_key(base_dir, schema_key);
+          else
+            full_schema_key = g_strdup(schema_key);
+        }
 
-	  g_free(full_schema_key);
-	}
+      if (!gconf_engine_associate_schema(conf, full_key, full_schema_key,  &error))
+        {
+          g_assert(error != NULL);
+          
+          g_printerr (_("WARNING: failed to associate schema `%s' with key `%s': %s\n"),
+                      full_schema_key, full_key, error->message);
+          g_error_free(error);
+        }
+
+      g_free(full_schema_key);
     }
 
   tmp = values;
@@ -2810,7 +2857,10 @@ set_values(GConfEngine* conf, const gchar* base_dir, const gchar* key, const cha
       GError* error;
 
       error = NULL;
-      gconf_engine_set(conf, full_key, value, &error);
+      if (!unload)
+        gconf_engine_set(conf, full_key, value, &error);
+      else
+        gconf_engine_unset(conf, full_key, &error);
       if (error != NULL)
         {
           g_printerr (_("Error setting value: %s\n"), error->message);
@@ -2825,7 +2875,7 @@ set_values(GConfEngine* conf, const gchar* base_dir, const gchar* key, const cha
 }
 
 static int
-process_entry(GConfEngine* conf, xmlNodePtr node, const gchar** base_dirs, const char* orig_base)
+process_entry(GConfEngine* conf, gboolean unload, xmlNodePtr node, const gchar** base_dirs, const char* orig_base)
 {
   xmlNodePtr iter;
   GSList* values = NULL;
@@ -2852,11 +2902,11 @@ process_entry(GConfEngine* conf, xmlNodePtr node, const gchar** base_dirs, const
   if (key && (values || schema_key))
     {
       if (!base_dirs)
-        set_values(conf, orig_base, key, schema_key, values);
+        set_values(conf, unload, orig_base, key, schema_key, values);
 
       else while (*base_dirs)
         {
-          set_values(conf, *base_dirs, key, schema_key, values);
+          set_values(conf, unload, *base_dirs, key, schema_key, values);
           ++base_dirs;
         }
     }
@@ -3321,20 +3371,16 @@ process_locale_info(xmlNodePtr node, SchemaInfo* info)
 }
 
 static int
-process_key_list(GConfEngine* conf, const gchar* schema_name, GSList* keylist)
+process_key_list(GConfEngine* conf, gboolean unload, const gchar* schema_name, GSList* keylist)
 {
   GSList* tmp;
-  GError* error = NULL;
-
-  if (makefile_uninstall_mode)
-    {
-      schema_name = NULL;
-    }
 
   tmp = keylist;
   while (tmp != NULL)
     {
-      if (!gconf_engine_associate_schema(conf, tmp->data, schema_name,  &error))
+      GError* error = NULL;
+
+      if (!gconf_engine_associate_schema(conf, tmp->data, !unload ? schema_name : NULL, &error))
         {
           g_assert(error != NULL);
           
@@ -3440,6 +3486,7 @@ hash_install_foreach(gpointer key, gpointer value, gpointer user_data)
 {
   struct {
     GConfEngine* conf;
+    gboolean unload;
     char* key;
   } *info;
   GConfSchema* schema;
@@ -3448,62 +3495,52 @@ hash_install_foreach(gpointer key, gpointer value, gpointer user_data)
   info = user_data;
   schema = value;
 
-  if (!gconf_engine_set_schema(info->conf, info->key, schema, &error))
+  if (!info->unload)
     {
-      g_assert(error != NULL);
+      if (!gconf_engine_set_schema(info->conf, info->key, schema, &error))
+	{
+	  g_assert(error != NULL);
 
-      g_printerr (_("WARNING: failed to install schema `%s' locale `%s': %s\n"),
-		  info->key, gconf_schema_get_locale(schema), error->message);
-      g_error_free(error);
-      error = NULL;
+	  g_printerr (_("WARNING: failed to install schema `%s' locale `%s': %s\n"),
+		      info->key, gconf_schema_get_locale(schema), error->message);
+	  g_error_free(error);
+	  error = NULL;
+	}
+      else
+	{
+	  g_assert(error == NULL);
+	  g_print (_("Installed schema `%s' for locale `%s'\n"),
+		   info->key, gconf_schema_get_locale(schema));
+	}
     }
   else
     {
-      g_assert(error == NULL);
-      g_print (_("Installed schema `%s' for locale `%s'\n"),
-	       info->key, gconf_schema_get_locale(schema));
-    }
+      if (!gconf_engine_unset (info->conf, info->key, &error))
+	{
+	  g_assert(error != NULL);
 
-  gconf_schema_free(schema);
-}
-
-static void
-hash_uninstall_foreach(gpointer key, gpointer value, gpointer user_data)
-{
-  struct {
-    GConfEngine* conf;
-    char* key;
-  } *info;
-  GConfSchema* schema;
-  GError* error = NULL;
-  
-  info = user_data;
-  schema = value;
-
-  if (!gconf_engine_unset (info->conf, info->key, &error))
-    {
-      g_assert(error != NULL);
-
-      g_printerr (_("WARNING: failed to uninstall schema `%s' locale `%s': %s\n"),
-		  info->key, gconf_schema_get_locale(schema), error->message);
-      g_error_free(error);
-      error = NULL;
-    }
-  else
-    {
-      g_assert(error == NULL);
-      g_print (_("Uninstalled schema `%s' from locale `%s'\n"),
-	       info->key, gconf_schema_get_locale(schema));
+	  g_printerr (_("WARNING: failed to uninstall schema `%s' locale `%s': %s\n"),
+		      info->key, gconf_schema_get_locale(schema), error->message);
+	  g_error_free(error);
+	  error = NULL;
+	}
+      else
+	{
+	  g_assert(error == NULL);
+	  g_print (_("Uninstalled schema `%s' from locale `%s'\n"),
+		   info->key, gconf_schema_get_locale(schema));
+	}
     }
 
   gconf_schema_free(schema);
 }
 
 static int
-process_schema(GConfEngine* conf, xmlNodePtr node)
+process_schema(GConfEngine* conf, gboolean unload, xmlNodePtr node)
 {
   struct {
     GConfEngine* conf;
+    gboolean unload;
     char* key;
   } hash_foreach_info;
   GHashTable* schemas_hash = NULL;
@@ -3519,14 +3556,12 @@ process_schema(GConfEngine* conf, xmlNodePtr node)
 
   if (schema_key != NULL)
     {
-      process_key_list(conf, schema_key, applyto_list);
+      process_key_list(conf, unload, schema_key, applyto_list);
 
       hash_foreach_info.conf = conf;
+      hash_foreach_info.unload = unload;
       hash_foreach_info.key = schema_key;
-      if (makefile_uninstall_mode)
-        g_hash_table_foreach(schemas_hash, hash_uninstall_foreach, &hash_foreach_info);
-      else 
-        g_hash_table_foreach(schemas_hash, hash_install_foreach, &hash_foreach_info);
+      g_hash_table_foreach(schemas_hash, hash_install_foreach, &hash_foreach_info);
     }
   else
     {
@@ -3550,7 +3585,7 @@ process_schema(GConfEngine* conf, xmlNodePtr node)
 }
 
 static int
-process_list(GConfEngine* conf, LoadType load_type, xmlNodePtr node, const gchar** base_dirs)
+process_list(GConfEngine* conf, LoadType load_type, gboolean unload, xmlNodePtr node, const gchar** base_dirs)
 {
 #define LOAD_TYPE_TO_LIST(t) ((t == LOAD_SCHEMA_FILE) ? "schemalist" : "entrylist")
 
@@ -3567,9 +3602,9 @@ process_list(GConfEngine* conf, LoadType load_type, xmlNodePtr node, const gchar
       if (iter->type == XML_ELEMENT_NODE)
         {
           if (load_type == LOAD_SCHEMA_FILE && strcmp(iter->name, "schema") == 0)
-            process_schema(conf, iter);
+            process_schema(conf, unload, iter);
           else if (load_type == LOAD_ENTRY_FILE && strcmp(iter->name, "entry") == 0)
-            process_entry(conf, iter, base_dirs, orig_base);
+            process_entry(conf, unload, iter, base_dirs, orig_base);
           else
             g_printerr (_("WARNING: node <%s> not understood below <%s>\n"),
 			iter->name, LOAD_TYPE_TO_LIST(load_type));
@@ -3587,7 +3622,7 @@ process_list(GConfEngine* conf, LoadType load_type, xmlNodePtr node, const gchar
 }
 
 static int
-do_load_file(GConfEngine* conf, LoadType load_type, const gchar* file, const gchar** base_dirs)
+do_load_file(GConfEngine* conf, LoadType load_type, gboolean unload, const gchar* file, const gchar** base_dirs)
 {
 #define LOAD_TYPE_TO_ROOT(t) ((t == LOAD_SCHEMA_FILE) ? "gconfschemafile" : "gconfentryfile")
 #define LOAD_TYPE_TO_LIST(t) ((t == LOAD_SCHEMA_FILE) ? "schemalist" : "entrylist")
@@ -3646,7 +3681,7 @@ do_load_file(GConfEngine* conf, LoadType load_type, const gchar* file, const gch
       if (iter->type == XML_ELEMENT_NODE)
         {
           if (strcmp(iter->name, LOAD_TYPE_TO_LIST(load_type)) == 0)
-            process_list(conf, load_type, iter, base_dirs);
+            process_list(conf, load_type, unload, iter, base_dirs);
           else
             g_printerr (_("WARNING: node <%s> below <%s> not understood\n"),
 			iter->name, LOAD_TYPE_TO_ROOT(load_type));
@@ -3672,7 +3707,7 @@ do_load_file(GConfEngine* conf, LoadType load_type, const gchar* file, const gch
 }
 
 static int
-do_makefile_install(GConfEngine* conf, const gchar** args)
+do_makefile_install(GConfEngine* conf, const gchar** args, gboolean unload)
 {
   GError* err = NULL;
   
@@ -3684,39 +3719,7 @@ do_makefile_install(GConfEngine* conf, const gchar** args)
 
   while (*args)
     {
-      if (do_load_file(conf, LOAD_SCHEMA_FILE, *args, NULL) != 0)
-        return 1;
-
-      ++args;
-    }
-
-  gconf_engine_suggest_sync(conf, &err);
-
-  if (err != NULL)
-    {
-      g_printerr (_("Error syncing config data: %s"),
-		  err->message);
-      g_error_free(err);
-      return 1;
-    }
-  
-  return 0;
-}
-
-static int
-do_makefile_uninstall(GConfEngine* conf, const gchar** args)
-{
-  GError* err = NULL;
-  
-  if (args == NULL)
-    {
-      g_printerr (_("Must specify some schema files to uninstall\n"));
-      return 1;
-    }
-
-  while (*args)
-    {
-      if (do_load_file(conf, LOAD_SCHEMA_FILE, *args, NULL) != 0)
+      if (do_load_file(conf, LOAD_SCHEMA_FILE, unload, *args, NULL) != 0)
         return 1;
 
       ++args;
@@ -3978,7 +3981,7 @@ do_get_default_source (const gchar** args)
   gchar buf[512];
   FILE *f;
 
-  /* Try with $sysconfdir/gconf/schema-install-source */
+  /* Try with $sysgconfdir/schema-install-source */
   f = fopen(GCONF_ETCDIR"/schema-install-source", "r");
 
   if (f != NULL)
